/**
 * Copyright (c) 2011-2012, Institute of Electronics and Computer Science
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of  conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * fs/block/dir.c -- directory operations
 */

#include <limits.h>
#include <stdint.h>
#include <string.h>

#include <fs/common.h>
#include <eeprom.h>
#include <sleep.h>
#include <lib/assert.h>
#include <lib/dprint.h>

#include "alloc.h"
#include "common.h"
#include "flash_access.h"

#include "dir.h"

#define MAX_DIR_ENTRIES (BLOCK_DATA_SIZE / sizeof(struct fsFileEntry))

/* File entry on flash */
struct fsFileEntry {
    char     name[FS_NAME_MAX];
    blk_t    first;
    fsOff_t  size;
    uint8_t  attr;
    uint16_t lastCrc;
} PACKED;

/* File entry attributes */
enum fileAttributes {
    ATTR_EMPTY = 1 << 0, /* Entry is empty, can be used */
    ATTR_VALID = 1 << 1, /* Entry is allocated and valid */
    ATTR_DIR   = 1 << 2,
};


/* Array of file control blocks; 1 extra for dir traversal */
struct fsFileControlBlock fcbs[FS_MAX_OPEN_FILES+1];
mos_mutex_t               fsFcbMutex;

static struct fsFileControlBlock *findFcb(blk_t dir, unsigned off)
{
    size_t i;

    for (i = 0; i < FS_MAX_OPEN_FILES+1; i++)
    {
        if (fcbs[i].id.dir == dir && fcbs[i].id.off == off)
            return fcbs + i;
    }

    return NULL;
}

/* Special version for searching for file parent directories */
static struct fsFileControlBlock *findFcbByFirst(blk_t first)
{
    size_t i;

    for (i = 0; i < FS_MAX_OPEN_FILES+1; i++)
    {
        if (fcbs[i].id.dir != BLOCK_INVAL && fcbs[i].id.off != -1 &&
            fcbs[i].first == first)
        {
            return fcbs + i;
        }
    }

    return NULL;
}

/* Allocate a file control block from array */
static inline struct fsFileControlBlock *allocFcb(void)
{
    struct fsFileControlBlock *res = findFcb(BLOCK_INVAL, -1);
    if (!res)
        fsSetError(FS_ERR_NOMEM);
    return res;
}

inline void freeFcb(struct fsFileControlBlock *fcb)
{
    fcb->id.dir = BLOCK_INVAL;
    fcb->id.off = -1;
}


static inline uint32_t entryAddr(unsigned pos)
{
    return pos * sizeof(struct fsFileEntry);
}

#define READ_ENTRY_FIELD_N(dir, pos, field, buf, len)                       \
    blkExtFlashRead(flashAddr(dir, entryAddr(pos)                           \
                                   + offsetof(struct fsFileEntry, field)),  \
                    buf, len);
#define READ_ENTRY_FIELD(dir, pos, field, buf)                              \
    READ_ENTRY_FIELD_N(dir, pos, field, &buf, sizeof(buf))
#define WRITE_ENTRY_FIELD_N(dir, pos, field, buf, len)                      \
    blkExtFlashWrite(flashAddr(dir, entryAddr(pos)                          \
                                    + offsetof(struct fsFileEntry, field)), \
                     buf, len);
#define WRITE_ENTRY_FIELD(dir, pos, field, buf)                             \
    WRITE_ENTRY_FIELD_N(dir, pos, field, &buf, sizeof(buf))

static inline void readEntry(blk_t dir, unsigned pos, struct fsFileEntry *buf)
{
    blkExtFlashRead(flashAddr(dir, entryAddr(pos)), buf, sizeof(struct fsFileEntry));
}

static inline void writeEntry(blk_t dir, unsigned pos, const struct fsFileEntry *buf)
{
    blkExtFlashWrite(flashAddr(dir, entryAddr(pos)), buf, sizeof(struct fsFileEntry));
}

static bool nextEntry(blk_t dir, unsigned *start, bool empty)
{
    while (++*start < MAX_DIR_ENTRIES)
    {
        uint8_t attr;
        READ_ENTRY_FIELD(dir, *start, attr, attr);
        if (empty ? attr & ATTR_EMPTY : !(attr & ATTR_EMPTY) && (attr & ATTR_VALID))
            return true;
        else if (!empty && attr & ATTR_EMPTY) /* End of valid entries */
            return false;
    }
    return false;
}

static inline bool nextValidEntry(blk_t dir, unsigned *start)
{
    return nextEntry(dir, start, false);
}

static inline bool nextEmptyEntry(blk_t dir, unsigned *start)
{
    return nextEntry(dir, start, true);
}

blk_t fsRootDir(void)
{
    blk_t res;
    eepromRead(EEPROM_OFFSET + offsetof(struct fsHeader, root),
               &res, sizeof(res));
    return res;
}

void *fsBlockOpenLastDir(void *parent)
{
    struct fsFileControlBlock *fcb, *par = parent;

    mos_mutex_lock(&fsFcbMutex);

    if (parent)
        fcb = findFcb(par->first, par->pos);
    else
        fcb = findFcb(BLOCK_INVAL, 0);
    if (fcb)
    {
        fsSetError(FS_ERR_OPEN);
        fcb = NULL;
    }
    else if ((fcb = allocFcb()))
    {
        if (par)
        {
            fcb->id.dir = par->first;
            fcb->id.off = par->pos;
            READ_ENTRY_FIELD(par->first, par->pos, first, fcb->first);
        }
        else
        {
            fcb->id.dir = BLOCK_INVAL;
            fcb->id.off = 0;
            fcb->first  = fsRootDir();
        }
        fcb->pos = -1;
    }

    mos_mutex_unlock(&fsFcbMutex);
    return fcb;
}

bool fsBlockReadDir(void * restrict dir, char * restrict name)
{
    struct fsFileControlBlock *d = dir;

    if (nextValidEntry(d->first, &d->pos))
    {
        READ_ENTRY_FIELD_N(d->first, d->pos, name, name, FS_NAME_MAX);
        name[FS_NAME_MAX] = '\0';
        return true;
    }
    else
        return false;
}

void  fsBlockRewindDir(void *dir)
{
    struct fsFileControlBlock *d = dir;

    d->pos = -1;
}

void fsBlockCloseDir(void *dir)
{
    mos_mutex_lock(&fsFcbMutex);
    freeFcb(dir);
    mos_mutex_unlock(&fsFcbMutex);
}


bool fsBlockStatLast(void * restrict dir, struct fsStat * restrict buf)
{
    struct fsFileControlBlock *d = dir;

    READ_ENTRY_FIELD(d->first, d->pos, size, buf->size);
    uint8_t a;
    READ_ENTRY_FIELD(d->first, d->pos, attr, a);
    buf->dir = a & ATTR_DIR;

    return true;
}


static void removeCurEntry(struct fsFileControlBlock *dir)
{
    uint8_t a;
    READ_ENTRY_FIELD(dir->first, dir->pos, attr, a);
    a &= ~ATTR_VALID;
    WRITE_ENTRY_FIELD(dir->first, dir->pos, attr, a);
}

bool fsBlockRemoveLast(void *dir)
{
    struct fsFileControlBlock *d = dir, *fcb;

    mos_mutex_lock(&fsFcbMutex);
    fcb = findFcb(d->first, d->pos);
    mos_mutex_unlock(&fsFcbMutex);
    if (fcb)
    {
        fsSetError(FS_ERR_OPEN);
        return false;
    }

    /* Walk the file and delete its blocks */
    fsOff_t size;
    blk_t   curr, next;
    READ_ENTRY_FIELD(d->first, d->pos, size, size);
    READ_ENTRY_FIELD(d->first, d->pos, first, curr);
    while (size)
    {
        next = fsBlockGetNext(curr);
        fsBlockFree(curr);
        curr = next;
        size -= MIN(BLOCK_DATA_SIZE, size);
    }

    removeCurEntry(d);

    return true;
}

bool fsBlockRenameLast(void * restrict dir, const char * restrict name)
{
    struct fsFileControlBlock *d = dir, *fcb;
    struct fsFcbId             save = { d->first, d->pos };

    mos_mutex_lock(&fsFcbMutex);
    fcb = findFcb(d->first, d->pos);
    mos_mutex_unlock(&fsFcbMutex);
    if (fcb)
    {
        fsSetError(FS_ERR_OPEN);
        return false;
    }

    struct fsFileEntry entry;
    readEntry(save.dir, save.off, &entry);

    removeCurEntry(d); /* FIXME: If nextEmptyEntry fails, state is altered */
    if (!nextEmptyEntry(d->first, &d->pos))
        return false;

    memset(entry.name, '\0', sizeof(entry.name));
    strncpy(entry.name, name, strlen(name));
    writeEntry(d->first, d->pos, &entry);

    return true;
}

bool fsBlockCreateAt(void * restrict parent, const char * restrict name,
                     bool dir)
{
    struct fsFileControlBlock *par = parent;
    struct fsFileEntry entry;

    if (!nextEmptyEntry(par->first, &par->pos))
        return false;

    if (dir)
    {
        if ((entry.first = fsBlockAllocate(fsRootDir())) == BLOCK_INVAL)
            return false;
        entry.size = BLOCK_DATA_SIZE;
    }
    else
    {
        entry.first = BLOCK_INVAL;
        entry.size  = 0;
    }

    memset(entry.name, '\0', sizeof(entry.name));
    strncpy(entry.name, name, strlen(name));
    entry.lastCrc = 0;
    entry.attr = ATTR_VALID;
    if (dir)
        entry.attr |= ATTR_DIR;

    writeEntry(par->first, par->pos, &entry);
    return true;
}


struct fsFileControlBlock *fsBlockOpenFcb(struct fsFileControlBlock *dir,
                                          bool write)
{
    struct fsFileControlBlock *res;

    mos_mutex_lock(&fsFcbMutex);

    res = findFcb(dir->first, dir->pos);
    if (res)
    {
        /* File is already open */
        if (write && res->openForWrite)
        {
            /* Can't have two writers */
            fsSetError(FS_ERR_OPEN);
            res = NULL;
        }
        else
        {
            res->refcount++;
            if (write)
                res->openForWrite = true;
        }
    }
    else
    {
        res = allocFcb();
        if (res)
        {
            res->id.dir       = dir->first;
            res->id.off       = dir->pos;
            res->refcount     = 1;
            res->openForWrite = write;

            READ_ENTRY_FIELD(dir->first, dir->pos, size,    res->size );
            READ_ENTRY_FIELD(dir->first, dir->pos, first,   res->first);
            READ_ENTRY_FIELD(dir->first, dir->pos, lastCrc, res->crc  );
        }
    }

    mos_mutex_unlock(&fsFcbMutex);
    return res;
}

void fsBlockSyncEntry(struct fsFileControlBlock *file)
{
    struct fsFileEntry entry;
    int                retries = 5;

    /*
     * If the directory is held open by another thread and we modify the
     * directory's state, it is possible that reading the directory will yield
     * incorrect or double file records. Therefore we have to wait until the
     * directory is closed. It can't be just a mutex lock operation on the
     * directory FCB, because the FCB is freed upon close and might be assigned
     * elsewhere. I would like having a global condition variable, on which
     * this function blocks and on which fsBlockCloseDir() signals. But, for
     * now, it's just polling (bad bad!).
     */
    mos_mutex_lock(&fsFcbMutex);
    while (findFcbByFirst(file->id.dir))
    {
        mos_mutex_unlock(&fsFcbMutex);

        if (!retries)
        {
            PRINTF("Could not sync file entry\n");
            return;
        }
        else
        {
            retries--;
            msleep(500);
            mos_mutex_lock(&fsFcbMutex);
        }
    }

    /* Read old entry */
    readEntry(file->id.dir, file->id.off, &entry);

    /* Mark the entry invalid */
    uint8_t attr = entry.attr & ~ATTR_VALID;
    WRITE_ENTRY_FIELD(file->id.dir, file->id.off, attr, attr);

    entry.first   = file->first;
    entry.size    = file->size;
    entry.lastCrc = file->crc;

    if (!nextEmptyEntry(file->id.dir, &file->id.off))
    {
        /* TODO */
        PRINTF("Directory space exhausted, expect file system failure\n");
        mos_mutex_unlock(&fsFcbMutex);
        return;
    }

    writeEntry(file->id.dir, file->id.off, &entry);
    mos_mutex_unlock(&fsFcbMutex);
}

void fsBlockCloseFcb(struct fsFileControlBlock *file)
{
    if (--file->refcount == 0)
    {
        fsBlockSyncEntry(file);
        mos_mutex_lock(&fsFcbMutex);
        freeFcb(file);
        mos_mutex_unlock(&fsFcbMutex);
    }
}
