/**
 * Copyright (c) 2011-2012, Institute of Electronics and Computer Science
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of  conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * fs/block/init.c -- subsystem initialization
 */

#include <hil/alarms.h>
#include <eeprom.h>
#include <lib/assert.h>

#include "alloc.h"
#include "common.h"
#include "dir.h"
#include "file.h"
#include "flash_access.h"

#include "init.h"

const struct fsOperations fsBlockOps = {
    .statLast   = fsBlockStatLast,   .createAt    = fsBlockCreateAt,
    .openLast   = fsBlockOpenLast,   .removeLast  = fsBlockRemoveLast,
    .renameLast = fsBlockRenameLast, .openLastDir = fsBlockOpenLastDir,
    .readDir    = fsBlockReadDir,    .closeDir    = fsBlockCloseDir,
    .read       = fsBlockRead,       .tell        = fsBlockTell,
    .seek       = fsBlockSeek,       .write       = fsBlockWrite,
    .flush      = fsBlockFlush,      .close       = fsBlockClose,
    .rewindDir  = fsBlockRewindDir
};

/* Flash sleep timer */
Alarm_t fsBlockFlashAlarm;

/* Process flash timer event */
void fsBlockAlarmCallback(void *data)
{
    (void)data;

    extFlashSleep(); /* TODO: I hope this is safe to call from interrupt */
}

#define MAGIC 0x8550 /* Chosen by fair dice rolls */

void fsBlockInit(void)
{
    size_t   i;
    uint16_t check;

    eepromRead(EEPROM_OFFSET + offsetof(struct fsHeader, magic),
               &check, sizeof(check));
    if (check != MAGIC)
    {
        /* There is no our signature. Let's initialize an empty filesystem. */

        /* Mark all blocks as available */
        fsBlockFreeAll();

        /* Allocate space for root directory */
        blk_t root = fsBlockAllocate(BLOCK_INVAL);
        ASSERT(root != BLOCK_INVAL);
        eepromWrite(EEPROM_OFFSET + offsetof(struct fsHeader, root),
                    &root, sizeof(root));

        /* Write magic */
        check = MAGIC;
        eepromWrite(EEPROM_OFFSET + offsetof(struct fsHeader, magic),
                    &check, sizeof(check));
    }

    mos_mutex_init(&fsFcbMutex);
    mos_mutex_init(&fsBlkTableMutex);
    mos_mutex_init(&fsBlkHandleMutex);

    alarmInit(&fsBlockFlashAlarm, fsBlockAlarmCallback, NULL);

    /* Initialize the file control block array */
    for (i = 0; i < FS_MAX_OPEN_FILES+1; i++)
    {
        mos_mutex_init(&fcbs[i].mutex);
        freeFcb(fcbs + i);
    }
}
