#include "stdfarmos.h"
#include "vlit.h"
#include <ctype.h>
#include <stdlib.h>

#include <simple_mhop.h>

#define MAX_NODE_COUNT  1

static void vlitRecv(uint8_t c);

static uint8_t vlitSeqnum;

static bool hasNewValues[MAX_NODE_COUNT];
static uint16_t nodeAddress[MAX_NODE_COUNT];
static struct vlit_values_t nodeData[MAX_NODE_COUNT];
static uint8_t nodeCount = 0;

// --------------------------------------

void vlitInit(void)
{
    serialInit(GATEWAY_USART_ID, GATEWAY_USART_BAUDRATE, 0);
    serialEnableRX(GATEWAY_USART_ID);
    serialEnableTX(GATEWAY_USART_ID);
    serialSetReceiveHandle(GATEWAY_USART_ID, vlitRecv);
}

// Note: this uses different polynom than in mos/lib/crc.h!
static uint16_t vlitCrcSum(volatile const void *data_, uint8_t length, uint16_t startCrc)
{
    uint16_t crc = startCrc;
    uint8_t i, pom, byte;
    volatile const uint8_t *data = (volatile const void *)data_;

#define CRC_POLYNOM 0xA001

    while (length--) {
        byte = *data;
        data++;
        for (i = 0; i < 8; i++) {
            pom = (byte & 0x01) ^ (crc & 0x0001);
            byte >>= 1;
            crc >>= 1;
            if (pom) {
                crc ^= CRC_POLYNOM;
            }
        }
    }
    return crc;
}

static inline void vlitSetHeader(struct umc_header_t * header, uint8_t length, uint8_t command, uint8_t sqn, uint16_t src)
{
    header->start = UMC_START_BYTE;
    header->len = sizeof(struct umc_header_t) - 2 + length;
    header->cmd = command;
    header->src = src;
    header->dst = 0x00; // FIXME ?
    header->sqn = sqn;
}

//
// We send data to gateway
//
void vlitSend(uint8_t cmd, void *data, uint8_t len, uint16_t sender)
{
    struct umc_header_t header;
    uint16_t crc;

    vlitSetHeader(&header, len, cmd, vlitSeqnum++, sender);
    crc = vlitCrcSum(data, len, vlitCrcSum(&header, sizeof(header), 0));

    serialSendData(GATEWAY_USART_ID, (uint8_t *)&header, sizeof(header));
    serialSendData(GATEWAY_USART_ID, (uint8_t *)data, len);
    serialSendData(GATEWAY_USART_ID, (uint8_t *)&crc, sizeof(crc));
}

//
// Gateway sends ACK to us
//
void vlitHandleReceiveDataReply(struct umc_header_t *header, uint8_t * data, uint8_t cnt)
{
//    PRINTF("got receive data reply\n");
    //
    // TODO?: check seqnum and check if code is 0 (OK) or 1 (error)
    //
}

//
// Gateway sends data request to us
//
void vlitHandleSendData(struct umc_header_t *header, uint8_t * data, uint8_t cnt)
{
    // PRINTF("got send data reply\n");
    // TODO: send the command out to sensor network
    // TODO: send ACK back to gateway (required by docs, but not by source code)

    uint16_t rfaddr;
    memcpy(&rfaddr, (const char*)data, 2);

    uint8_t i;
    for (i = 0; i < nodeCount; i++) {
        if (nodeAddress[i] == rfaddr) {
            if (hasNewValues[i] == true) {
                // send to the gateway
                static vlit_data_t data;
                data.metadata.rf_adr = rfaddr;
                data.metadata.app_len = sizeof(vlit_values_t);
                data.metadata.app_cmd = 0;

                data.values.soil_temp = nodeData[i].soil_temp;
                data.values.soil_humi = nodeData[i].soil_humi;
                data.values.air_temp = nodeData[i].air_temp;
                data.values.air_humi = nodeData[i].air_humi;
                data.values.bat_volt = nodeData[i].bat_volt;
                data.values.rssi = nodeData[i].rssi;

                vlitSend(UMC_CMD_RECEIVE_DATA, &data, sizeof(data), rfaddr);
                hasNewValues[i] = false;
            }
            break;
        }
    }
}

static void vlitRecv(uint8_t c)
{
    enum states {STATE_OFF = 0, STATE_HEADER, STATE_BODY, STATE_CRC};

    static uint8_t state = STATE_OFF;
    static uint8_t cnt = 0, i = 0;
    static uint8_t data[32];
    static uint8_t crc[2];
    static struct umc_header_t header;

    switch (state) {
    case STATE_OFF:
        if (c == UMC_START_BYTE) {
            state = STATE_HEADER;
            cnt = sizeof(struct umc_header_t);
            data[0] = c;
            i = 1;
        }
        break;

    case STATE_HEADER:
        data[i] = c;
        i++;
        if (i == cnt) {
            memcpy((void *)&header, data, sizeof(struct umc_header_t));
            if ((header.len - (sizeof(struct umc_header_t) - 2)) > 0) {
                state = STATE_BODY;
                i = 0;
                cnt = header.len - (sizeof(struct umc_header_t) - 2);
            }
            else {
                state = STATE_CRC;
            }
        }
        break;

    case STATE_BODY:
        data[i] = c;
        i++;
        if (i == cnt) {
            i = 0;
            state = STATE_CRC;
        }
        break;

    case STATE_CRC:
        crc[i] = c;
        i++;
        if (i == 2) {
            state = STATE_OFF;

            switch (header.cmd) {
            case UMC_CMD_SEND_DATA:
                vlitHandleSendData(&header, data, cnt);
                break;
            case UMC_CMD_RECEIVE_DATA_REPLY:
                vlitHandleReceiveDataReply(&header, data, cnt);
                break;
            case UMC_CMD_NETCMD:
                // Receive a command from the gateway and relay it to the network

                if (cnt == 4 && !strncmp("test", (const char *)data, 4))
                {
                    // Command processing example
                    vlitSend(UCM_CMD_NETCMD_ANSWER, "OK", 2, 42);
                }
                else if (cnt == 4 && !strncmp("seed", (const char *)data, 4))
                {
                    static const char *s = "Broadcasting seed";
                    vlitSend(UCM_CMD_NETCMD_ANSWER, (char *)s, strlen(s), 1);

                    broadcastSeed(NULL);
                }
                else if (cnt >= 7 && !strncmp("period=", (const char *)data, 7))
                {
                    data[cnt] = '\0';

                    int res = atoi((char *)(data+7));
                    if (res < 0 || 360 < res)
                    {
                        static const char *s = "Invalid number of minutes";
                        vlitSend(UCM_CMD_NETCMD_ANSWER, (char *)s, strlen(s), 1);
                    }
                    else
                    {
                        static const char *s = "OK; type \"netcmd 1 seed\" to force send.";
                        vlitSend(UCM_CMD_NETCMD_ANSWER, (char *)s, strlen(s), 1);

                        extern uint16_t periodFromGW;

                        periodFromGW = res;
                    }
                }
                break;
            default:
                // PRINTF("VLIT: unexpected cmd %#02x\n", header.cmd);
                break;
            }
        }
        break;

    }
}

// Set node data so it could be sent to GW by request
void setNodeData(struct vlit_values_t *data, uint16_t src)
{
    uint8_t i;
    bool isFound = false;

    for (i = 0; i < nodeCount; i++) {
        if (nodeAddress[i] == src) {
            isFound = true;
            break;
        }
    }

    if (!isFound) {
        i = nodeCount;
        nodeCount = nodeCount + 1;
    }
    nodeData[i].soil_temp = data->soil_temp;
    nodeData[i].soil_humi = data->soil_humi;
    nodeData[i].air_temp = data->air_temp;
    nodeData[i].air_humi = data->air_humi;
    nodeData[i].bat_volt = data->bat_volt;
    nodeData[i].rssi = data->rssi;
    nodeAddress[i] = src;
    hasNewValues[i] = true;
}
