/*
 * Copyright (c) 2013, Institute of Electronics and Computer Science
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of  conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _PLATFORM_MIIMOTE_H_
#define _PLATFORM_MIIMOTE_H_

#include <msp430/msp430_clock.h>
#include <msp430/msp430_timers.h>
#include <msp430/msp430_int.h>
#include <msp430/msp430_adc10.h>
#include <msp430/msp430_usci.h>

#include "flash.h"

#include <delay.h>
#include <digital.h>
#include <humidity.h>
#include <cc1101_pins.h>

void msleep(uint16_t); // Cannot include <sleep.h> here

#define RADIO_CHIP RADIO_CHIP_CC1101
#define SNUM_CHIP  SNUM_DS2411

//===========================================================
// Functions
//===========================================================

void initPlatform(void);

//===========================================================
// Data types and constants
//===========================================================

#define DS2411_PORT 4
#define DS2411_PIN  2

#define DS2411_VCC_PORT 4
#define DS2411_VCC_PIN  7

#define SDI12_TX_PORT 3
#define SDI12_TX_PIN  4

#define SDI12_RX_PORT 3
#define SDI12_RX_PIN  5

#define SDI12_NO_INVERT_VALUES

#define ADC_LIGHT_PHOTOSYNTHETIC ADC_LIGHT_CHANNEL

// ADC channels
enum {
    ADC_RADIATION_CURRENT = 0,  // Solar radiation/charging current
    ADC_AIR_TEMP_HUMIDITY = 1,  // Air temperature/humidity
    ADC_LIGHT_CHANNEL     = 6,  // Light sensor
    ADC_AIR_PRESSURE      = 7,  // Barometric pressure
    ADC_SOIL_MOISTURE     = 12, // Soil moisture
    ADC_BATTERY_VOLTAGE   = 13, // Battery voltage
    ADC_RESERVED          = 14, // Reserved ADC channel (ADC14 on the schematic)
    ADC_WIND_DIR          = 15, // Wind direction
};

// a synonym, to be compatible with Farmmote
#define ADC_SOIL_HUMIDITY ADC_SOIL_MOISTURE

// Enable selected ADC channel
static inline void platformEnableADC(int ch)
{
    if (ch < 8)
    {
        ADC10AE0 |= BV(ch);
    }
    else
    {
        ADC10AE1 |= BV(ch - 8);
    }
}


// Pin setup

#define PLATFORM_SENSORS_VOLTAGE_PORT          2
#define PLATFORM_SENSORS_VOLTAGE_PIN           3

#define PLATFORM_PERIPHERALS_VOLTAGE_PORT      1
#define PLATFORM_PERIPHERALS_VOLTAGE_PIN       6

#define PLATFORM_SDI12_ENABLE_PORT             4
#define PLATFORM_SDI12_ENABLE_PIN              0

#define PLATFORM_UART_SDI12_SWITCH_PORT        1
#define PLATFORM_UART_SDI12_SWITCH_PIN         0

#define PLATFORM_RADIATION_CURRENT_SWITCH_PORT 1
#define PLATFORM_RADIATION_CURRENT_SWITCH_PIN  4

#define PLATFORM_TEMP_HUMIDITY_SWITCH_PORT     4
#define PLATFORM_TEMP_HUMIDITY_SWITCH_PIN      1

#define PLATFORM_WINDSPEED_PORT                1
#define PLATFORM_WINDSPEED_PIN                 1

// Will also enable/disable UART comms
static inline void platformEnableSensors(void)
{
    // TODO: do not allow to enable (enter panic?) when current is too low

    // Turn on voltage
    pinAsOutput(PLATFORM_SENSORS_VOLTAGE_PORT, PLATFORM_SENSORS_VOLTAGE_PIN);
    pinClear(PLATFORM_SENSORS_VOLTAGE_PORT, PLATFORM_SENSORS_VOLTAGE_PIN);

    // Select temperature -- switch this early to not to wait for additional time
    pinClear(PLATFORM_TEMP_HUMIDITY_SWITCH_PORT, PLATFORM_TEMP_HUMIDITY_SWITCH_PIN);

    pinAsFunction(SDI12_TX_PORT, SDI12_TX_PIN);
    pinAsOutput(SDI12_TX_PORT, SDI12_TX_PIN);
    pinAsOutput(PLATFORM_TEMP_HUMIDITY_SWITCH_PORT, PLATFORM_TEMP_HUMIDITY_SWITCH_PIN);
    pinAsOutput(PLATFORM_SDI12_ENABLE_PORT,  PLATFORM_SDI12_ENABLE_PIN);

#ifdef USE_HUMIDITY
    // Turn on external temp/humidity sensor
    humidityOn();
#endif

    // For air temp/humidity sensor (150 ms), air pressure sensor (600 ms),
    // light sensor (1000 ms) init; hopefully also enough for external
    // temp/humidity sensor (the datasheet says 5-30 s, too long to wait)
    msleep(1050);
}


#define XREN(port) P##port##REN
#define REN(port)  XREN(port)

static inline void platformDisableSensors(void)
{
#ifdef USE_HUMIDITY
    humidityOff();
#endif
    pinAsInput(PLATFORM_SENSORS_VOLTAGE_PORT, PLATFORM_SENSORS_VOLTAGE_PIN);
    REN(PLATFORM_SENSORS_VOLTAGE_PORT) &= ~BV(PLATFORM_SENSORS_VOLTAGE_PIN);

    pinAsData(SDI12_TX_PORT, SDI12_TX_PIN);
    pinAsOutput(SDI12_TX_PORT, SDI12_TX_PIN);
    pinClear(SDI12_TX_PORT, SDI12_TX_PIN);
    pinAsInput(PLATFORM_TEMP_HUMIDITY_SWITCH_PORT, PLATFORM_TEMP_HUMIDITY_SWITCH_PIN);
    pinAsInput(PLATFORM_SDI12_ENABLE_PORT,  PLATFORM_SDI12_ENABLE_PIN);
}

static inline void platformEnablePeripherals(void)
{
    pinAsOutput(PLATFORM_PERIPHERALS_VOLTAGE_PORT, PLATFORM_PERIPHERALS_VOLTAGE_PIN);
    pinClear(PLATFORM_PERIPHERALS_VOLTAGE_PORT, PLATFORM_PERIPHERALS_VOLTAGE_PIN);

    pinAsData(CC1101_CSN_PORT, CC1101_CSN_PIN);
    pinAsFunction(CC1101_SI_PORT,  CC1101_SI_PIN);
    pinAsFunction(CC1101_SO_PORT,  CC1101_SO_PIN);
    pinAsFunction(CC1101_CLK_PORT, CC1101_CLK_PIN);
    pinAsOutput(CC1101_CSN_PORT, CC1101_CSN_PIN);

    pinAsOutput(1, 7); // LED
}

static inline void platformDisablePeripherals(void)
{
    // The following two lines somehow mess up the soil temperature sensor
    pinAsOutput(PLATFORM_PERIPHERALS_VOLTAGE_PORT, PLATFORM_PERIPHERALS_VOLTAGE_PIN);
    pinSet(PLATFORM_PERIPHERALS_VOLTAGE_PORT, PLATFORM_PERIPHERALS_VOLTAGE_PIN);

    pinAsData(CC1101_CSN_PORT, CC1101_CSN_PIN);
    pinAsData(CC1101_SI_PORT,  CC1101_SI_PIN);
    pinAsData(CC1101_SO_PORT,  CC1101_SO_PIN);
    pinAsData(CC1101_CLK_PORT, CC1101_CLK_PIN);
    pinAsInput(CC1101_CSN_PORT, CC1101_CSN_PIN);
    pinAsInput(CC1101_SI_PORT,  CC1101_SI_PIN);
    pinAsInput(CC1101_SO_PORT,  CC1101_SO_PIN);
    pinAsInput(CC1101_CLK_PORT, CC1101_CLK_PIN);

    pinAsInput(1, 7); // LED
}

static inline void platformSelectUART(void)
{
    pinClear(PLATFORM_SDI12_ENABLE_PORT, PLATFORM_SDI12_ENABLE_PIN);
    pinSet(PLATFORM_UART_SDI12_SWITCH_PORT, PLATFORM_UART_SDI12_SWITCH_PIN);
    pinAsFunction(SDI12_RX_PORT, SDI12_RX_PIN);
    pinAsFunction(SDI12_TX_PORT, SDI12_TX_PIN);
}

static inline void platformSelectSDI12(void)
{
    pinSet(PLATFORM_SDI12_ENABLE_PORT, PLATFORM_SDI12_ENABLE_PIN);
    pinClear(PLATFORM_UART_SDI12_SWITCH_PORT, PLATFORM_UART_SDI12_SWITCH_PIN);
    pinAsData(SDI12_RX_PORT, SDI12_RX_PIN);
    pinAsData(SDI12_TX_PORT, SDI12_TX_PIN);
}

static inline void platformSelectTemp(void)
{
    pinClear(PLATFORM_TEMP_HUMIDITY_SWITCH_PORT, PLATFORM_TEMP_HUMIDITY_SWITCH_PIN);
    msleep(750); // To stabilize ADC input voltage
}

static inline void platformSelectHumidity(void)
{
    pinSet(PLATFORM_TEMP_HUMIDITY_SWITCH_PORT, PLATFORM_TEMP_HUMIDITY_SWITCH_PIN);
    msleep(750); // To stabilize ADC input voltage
}

static inline void platformSelectCurrent(void)
{
    pinClear(PLATFORM_RADIATION_CURRENT_SWITCH_PORT, PLATFORM_RADIATION_CURRENT_SWITCH_PIN);
    msleep(15);
}

static inline void platformSelectRadiation(void)
{
    pinSet(PLATFORM_RADIATION_CURRENT_SWITCH_PORT, PLATFORM_RADIATION_CURRENT_SWITCH_PIN);
    msleep(15);
}

// external soil temperature sensor
#define DS18B20_PORT 2
#define DS18B20_PIN  4

// serial pins, for sw serial (TODO: check if correct!)
#define UART0_TX_PORT 3
#define UART0_TX_PIN  4
#define UART0_RX_PORT 3
#define UART0_RX_PIN  5

#define UART1_TX_PORT 3
#define UART1_TX_PIN  6
#define UART1_RX_PORT 3
#define UART1_RX_PIN  7

#endif
